#include "LibSip_SIPTypesAndValues.hh"
#include "LibSip_Common.hh"

#include "sip_codec_request.hh"
#include "loggers.hh"

int sip_codec_request::encode (const LibSip__SIPTypesAndValues::Request& msg, OCTETSTRING& data)
{
  loggers::get_instance().log_msg(">>> sip_codec_request::encode: ", (const Base_Type&)msg);

  osip_message_t* _sip_message;
  ::parser_init();
  int result = ::osip_message_init(&_sip_message);
  if (result != 0) {
    loggers::get_instance().warning("sip_codec_request::encode: Failed to initialise internal data structures");
    return -1;
  }
  
  char* buffer = nullptr;
  size_t length = 0;
  result = ::osip_message_to_str(_sip_message, &buffer, &length);
  if (result != 0) {
    loggers::get_instance().warning("sip_codec_request::encode: Failed to encode data structures");
    ::osip_message_free(_sip_message);
    return -1;
  }
  ::osip_message_free(_sip_message);
  data = OCTETSTRING(length, (const unsigned char*)buffer);

  loggers::get_instance().log_msg("<<< sip_codec_request::encode: data=", data);
  return result;
}

int sip_codec_request::decode (const OCTETSTRING& data, LibSip__SIPTypesAndValues::Request& msg, params* p_params)
{
  loggers::get_instance().log_msg(">>> sip_codec_request::decode: data=", data);

  // Sanity checks
  if (data.lengthof() == 0) {
    loggers::get_instance().warning("sip_codec_request::decode: Wrong parameters");
    return -1;
  }

  osip_message_t* _sip_message;
  ::parser_init();
  int result = ::osip_message_init(&_sip_message);
  if (result != 0) {
    loggers::get_instance().warning("sip_codec_request::decode: Failed to initialise internal data structures");
    return -1;
  }
  result = ::osip_message_parse(_sip_message, (const char*)static_cast<const unsigned char*>(data), data.lengthof());
  if (result != 0) {
    loggers::get_instance().warning("sip_codec_request::decode: Failed to parse SIP message");
    return -1;
  }

  // Fill RequestLine
  fill_request_line(_sip_message, msg);

  // Fill Headers
  fill_headers(_sip_message, msg);

  // Fill MessageBody
  fill_message_body(_sip_message, msg);
  
  // Fill Payload
  fill_payload(_sip_message, msg);
  
  ::osip_message_free(_sip_message);
  loggers::get_instance().log_msg("sip_codec_request::decode: ", msg);
  
  return 0;
}

void sip_codec_request::fill_headers(const osip_message_t* p_sip_message, LibSip__SIPTypesAndValues::Request& p_request)
{
  loggers::get_instance().log(">>> sip_codec_request::fill_headers");
  
  LibSip__SIPTypesAndValues::MessageHeader headers;
  // Fill mandatory fields
  // From
  LibSip__SIPTypesAndValues::From from_header;
  fill_from_header(::osip_message_get_from(p_sip_message), from_header);
  if (from_header.is_value()) {
    headers.fromField() = from_header;
  } else {
    headers.fromField().set_to_omit();
  }
  // To
  LibSip__SIPTypesAndValues::To to_header;
  fill_to_header(::osip_message_get_to(p_sip_message), to_header);
  if (to_header.is_value()) {
    headers.toField() = to_header;
  } else {
    headers.toField().set_to_omit();
  }
  // Via
  LibSip__SIPTypesAndValues::Via via_header;
  fill_via_header(p_sip_message, via_header);
  if (via_header.is_value()) {
    headers.via() = via_header;
  } else {
    headers.via().set_to_omit();
  }

  // Fill Optional fields
  LibSip__SIPTypesAndValues::Accept accept_header;
  fill_accept_header(p_sip_message, accept_header);
  if (accept_header.is_value()) {
    headers.accept() = accept_header;
  } else {
    headers.accept().set_to_omit();
  }

  LibSip__SIPTypesAndValues::AcceptContact accept_contact_header;
  fill_accept_contact_header(p_sip_message, accept_contact_header);
  if (accept_contact_header.is_value()) {
    headers.acceptContact() = accept_contact_header;
  } else {
    headers.acceptContact().set_to_omit();
  }
  
  LibSip__SIPTypesAndValues::AcceptEncoding accept_encoding_header;
  fill_accept_encoding_header(p_sip_message, accept_encoding_header);
  if (accept_encoding_header.is_value()) {
    headers.acceptEncoding() = accept_encoding_header;
  } else {
    headers.acceptEncoding().set_to_omit();
  }

  LibSip__SIPTypesAndValues::AcceptLanguage accept_language_header;
  fill_accept_language_header(p_sip_message, accept_language_header);
  if (accept_language_header.is_value()) {
    headers.acceptLanguage() = accept_language_header;
  } else {
    headers.acceptLanguage().set_to_omit();
  }

  LibSip__SIPTypesAndValues::AlertInfo alert_info_header;
  fill_alert_info_header(p_sip_message, alert_info_header);
  if (alert_info_header.is_value()) {
    headers.alertInfo() = alert_info_header;
  } else {
    headers.alertInfo().set_to_omit();
  }

  LibSip__SIPTypesAndValues::Allow allow_header;
  fill_allow_header(p_sip_message, allow_header);
  if (allow_header.is_value()) {
    headers.allow() = allow_header;
  } else {
    headers.allow().set_to_omit();
  }

  LibSip__SIPTypesAndValues::AllowEvents allow_events_header;
  fill_allow_events_header(p_sip_message, allow_events_header);
  if (allow_events_header.is_value()) {
    headers.allowEvents() = allow_events_header;
  } else {
    headers.allowEvents().set_to_omit();
  }

  LibSip__SIPTypesAndValues::Authorization authorization_header;
  fill_authorization_header(p_sip_message, authorization_header);
  if (authorization_header.is_value()) {
    headers.authorization() = authorization_header;
  } else {
    headers.authorization().set_to_omit();
  }
  headers.authenticationInfo().set_to_omit();
  headers.callId().set_to_omit();
  headers.callInfo().set_to_omit();
  headers.contact().set_to_omit();
  headers.contentDisposition().set_to_omit();
  headers.contentEncoding().set_to_omit();
  headers.contentLanguage().set_to_omit();
  headers.contentLength().set_to_omit();
  headers.contentType().set_to_omit();
  headers.cSeq().set_to_omit();
  headers.date().set_to_omit();
  headers.errorInfo().set_to_omit();
  headers.event().set_to_omit();
  headers.expires().set_to_omit();
  headers.featureCaps().set_to_omit();
  headers.geolocation().set_to_omit();
  headers.geolocationRouting().set_to_omit();
  headers.historyInfo().set_to_omit();
  headers.infoPackage().set_to_omit();
  headers.inReplyTo().set_to_omit();
  headers.maxForwards().set_to_omit();
  headers.mimeVersion().set_to_omit();
  headers.minExpires().set_to_omit();
  headers.minSE().set_to_omit();
  headers.organization().set_to_omit();
  headers.pAccessNetworkInfo().set_to_omit();
  headers.pAssertedID().set_to_omit();
  headers.pAssertedService().set_to_omit();
  headers.pAssociatedURI().set_to_omit();
  headers.path().set_to_omit();
  headers.pCalledPartyID().set_to_omit();
  headers.pChargingFunctionAddresses().set_to_omit();
  headers.pChargingVector().set_to_omit();
  headers.pEarlyMedia().set_to_omit();
  headers.pMediaAuthorization().set_to_omit();
  headers.pPreferredID().set_to_omit();
  headers.pPreferredService().set_to_omit();
  headers.priority().set_to_omit();
  headers.privacy().set_to_omit();
  headers.proxyAuthenticate().set_to_omit();
  headers.proxyAuthorization().set_to_omit();
  headers.proxyRequire().set_to_omit();
  headers.pVisitedNetworkID().set_to_omit();
  headers.rAck().set_to_omit();
  headers.rSeq().set_to_omit();
  headers.reason().set_to_omit();
  headers.recordRoute().set_to_omit();
  headers.recvInfo().set_to_omit();
  headers.requestDisposition().set_to_omit();
  headers.referredBy().set_to_omit();
  headers.referTo().set_to_omit();
  headers.referSub().set_to_omit();
  headers.replaces().set_to_omit();
  headers.replyTo().set_to_omit();
  headers.require().set_to_omit();
  headers.retryAfter().set_to_omit();
  headers.route().set_to_omit();
  headers.securityClient().set_to_omit();
  headers.securityServer().set_to_omit();
  headers.securityVerify().set_to_omit();
  headers.server().set_to_omit();
  headers.serviceRoute().set_to_omit();
  headers.sessionExpires().set_to_omit();
  headers.sessionId().set_to_omit();
  headers.sipETag().set_to_omit();
  headers.sipIfMatch().set_to_omit();
  headers.subject().set_to_omit();
  headers.subscriptionState().set_to_omit();
  headers.supported().set_to_omit();
  headers.timestamp().set_to_omit();
  headers.unsupported().set_to_omit();
  headers.userToUser().set_to_omit();
  headers.userAgent().set_to_omit();
  headers.warning().set_to_omit();
  headers.wwwAuthenticate().set_to_omit();
  headers.resourcePriority().set_to_omit();
  headers.answerMode().set_to_omit();
  headers.privAnswerMode().set_to_omit();
  headers.targetDialog().set_to_omit();
  headers.pAnswerState().set_to_omit();
  headers.undefinedHeader__List().set_to_omit();
  
  p_request.msgHeader() = headers;
  
  loggers::get_instance().log_msg("sip_codec_request::fill_headers: ", p_request);
} // End of method fill_headers

void sip_codec_request::fill_message_body(const osip_message_t* p_sip_message, LibSip__SIPTypesAndValues::Request& p_request)
{
  loggers::get_instance().log(">>> sip_codec_request::fill_message_body");

  p_request.messageBody().set_to_omit();
} // End of method fill_message_body

void sip_codec_request::fill_payload(const osip_message_t* p_sip_message, LibSip__SIPTypesAndValues::Request& p_request)
{
  loggers::get_instance().log(">>> sip_codec_request::fill_payload");

  p_request.payload().set_to_omit();
  
} // End of method fill_payload

void sip_codec_request::fill_semi_colon_params(const osip_list_t& p_sip_list, OPTIONAL<LibSip__Common::SemicolonParam__List>& p_list) {
  loggers::get_instance().log(">>> sip_codec_request::fill_semi_colon_params");

  if (::osip_list_size(&p_sip_list) != 0) {
    LibSip__Common::SemicolonParam__List l;
    for (int i = 0; i < ::osip_list_size(&p_sip_list); i++) {
      osip_generic_param_t* p = (osip_generic_param_t*)::osip_list_get(&p_sip_list, i);
      loggers::get_instance().log("sip_codec_request::fill_semi_colon_params: After cast: %p", p);
      CHARSTRING name(p->gname);
      if (p->gvalue == nullptr) {
        l[i] = LibSip__Common::GenericParam(name, OPTIONAL<LibSip__Common::GenValue>(OMIT_VALUE));
      } else {
        LibSip__Common::GenValue g;
        g.tokenOrHost() = CHARSTRING(p->gvalue);
        l[i] = LibSip__Common::GenericParam(name, g);
      }
    } // End of 'for' statement
    p_list = OPTIONAL<LibSip__Common::SemicolonParam__List>(l);
  } else {
    p_list.set_to_omit();
  }

  loggers::get_instance().log_msg("<<< sip_codec_request::fill_semi_colon_params: ", p_list);
} // End of method fill_semi_colon_params

void sip_codec_request::fill_ampersand_params(const osip_list_t& p_sip_list, OPTIONAL<LibSip__Common::AmpersandParam__List>& p_list) {
  loggers::get_instance().log(">>> sip_codec_request::fill_ampersand_params");

  if (::osip_list_size(&p_sip_list) != 0) {
    LibSip__Common::AmpersandParam__List l;
    for (int i = 0; i < ::osip_list_size(&p_sip_list); i++) {
      osip_generic_param_t* p = (osip_generic_param_t*)::osip_list_get(&p_sip_list, i);
      loggers::get_instance().log("sip_codec_request::fill_ampersand_params: After cast: %p", p);
      CHARSTRING name(p->gname);
      if (p->gvalue == nullptr) {
        l[i] = LibSip__Common::GenericParam(name, OPTIONAL<LibSip__Common::GenValue>(OMIT_VALUE));
      } else {
        LibSip__Common::GenValue g;
        g.tokenOrHost() = CHARSTRING(p->gvalue);
        l[i] = LibSip__Common::GenericParam(name, g);
      }
    } // End of 'for' statement
    p_list = OPTIONAL<LibSip__Common::AmpersandParam__List>(l);
  } else {
    p_list.set_to_omit();
  }

  loggers::get_instance().log_msg("<<< sip_codec_request::fill_ampersand_params: ", p_list);
} // End of method fill_ampersand_params

void sip_codec_request::fill_host_port(const char* p_host, const char* p_port, LibSip__SIPTypesAndValues::HostPort& p_host_port) {
  loggers::get_instance().log(">>> sip_codec_request::fill_host_port");
  
  if (p_host == nullptr) {
    p_host_port.host().set_to_omit();
  } else {
    p_host_port.host() = OPTIONAL<CHARSTRING>(CHARSTRING(p_host));
  }
  OPTIONAL<INTEGER> port;
  if (p_port == nullptr) {
    p_host_port.portField().set_to_omit();
  } else {
    p_host_port.portField() = OPTIONAL<INTEGER>(INTEGER(std::stoi(p_port)));
  }

  loggers::get_instance().log_msg("<<< sip_codec_request::fill_host_port: ", p_host_port);
} // End of method fill_host_port"

void sip_codec_request::fill_request_line(const osip_message_t* p_sip_message, LibSip__SIPTypesAndValues::Request& p_request)
{
  loggers::get_instance().log(">>> sip_codec_request::fill_request_line");
  
  LibSip__SIPTypesAndValues::RequestLine request_line;
  std::string str(::osip_message_get_method(p_sip_message));
  str += "_E";
  request_line.method() = LibSip__SIPTypesAndValues::Method(LibSip__SIPTypesAndValues::Method::str_to_enum(str.c_str()));
  loggers::get_instance().log_msg("sip_codec_request::fill_request_line: Method: ", request_line.method());
  LibSip__SIPTypesAndValues::SipUrl uri;
  fill_uri(uri, ::osip_message_get_uri(p_sip_message));
  loggers::get_instance().log_msg("sip_codec_request::fill_request_line: SipUrl: ", uri);
  request_line.requestUri() = uri;
  request_line.sipVersion() = CHARSTRING(::osip_message_get_version(p_sip_message));
  p_request.requestLine() = request_line;
  // FIXME To be continued

  loggers::get_instance().log_msg("<<< sip_codec_request::fill_request_line: ", p_request);
}

void sip_codec_request::fill_uri(LibSip__SIPTypesAndValues::SipUrl& p_sip_url, const osip_uri_t* p_uri) {
  loggers::get_instance().log(">>> sip_codec_request::fill_uri");
  
  // Scheme
  LibSip__SIPTypesAndValues::UriComponents uri_components;
  if (::osip_uri_get_scheme((osip_uri_t*)p_uri) != nullptr) {
    p_sip_url.scheme() = CHARSTRING(::osip_uri_get_scheme((osip_uri_t*)p_uri));
    loggers::get_instance().log_msg("sip_codec_request::fill_uri: scheme", p_sip_url.scheme());
    // SipUriComponents
    if (strcmp(::osip_uri_get_scheme((osip_uri_t*)p_uri), "sip") == 0) {
      // User Info
      OPTIONAL<LibSip__SIPTypesAndValues::UserInfo> user_info;
      if (::osip_uri_get_username((osip_uri_t*)p_uri) == nullptr) {
        user_info.set_to_omit();
      } else {
        OPTIONAL<CHARSTRING> password;
        if (::osip_uri_get_password((osip_uri_t*)p_uri) == nullptr) {
          password.set_to_omit();
        } else {
          password = CHARSTRING(::osip_uri_get_password((osip_uri_t*)p_uri));
        }
        LibSip__SIPTypesAndValues::UserInfo u(CHARSTRING(::osip_uri_get_username((osip_uri_t*)p_uri)), password);
        user_info = OPTIONAL<LibSip__SIPTypesAndValues::UserInfo>(u);
      }
      loggers::get_instance().log_msg("sip_codec_request::fill_uri: user_info: ", user_info);
      // HostPort
      LibSip__SIPTypesAndValues::HostPort host_port;
      fill_host_port(::osip_uri_get_host((osip_uri_t*)p_uri), ::osip_uri_get_port((osip_uri_t*)p_uri), host_port);
      loggers::get_instance().log_msg("sip_codec_request::fill_uri: host_port: ", host_port);
      uri_components.sip() = LibSip__SIPTypesAndValues::SipUriComponents(user_info, host_port);
    } else if (strcmp(::osip_uri_get_scheme((osip_uri_t*)p_uri), "tel") == 0) {
      // FIXME uri_components.tel() = LibSip__SIPTypesAndValues::TelUriComponents(user_info, host_port);
    } else if (strcmp(::osip_uri_get_scheme((osip_uri_t*)p_uri), "urn") == 0) {
      // FIXME uri_components.urn() = LibSip__SIPTypesAndValues::UrnUriComponents(user_info, host_port);
    } else {
      char *buffer = nullptr;
      ::osip_uri_to_str_canonical((osip_uri_t*)p_uri, &buffer);
      uri_components.other() = CHARSTRING(buffer);
      osip_free(buffer); // Macro
    }
  } else {
    char *buffer = nullptr;
    ::osip_uri_to_str_canonical((osip_uri_t*)p_uri, &buffer);
    uri_components.other() = CHARSTRING(buffer);
    osip_free(buffer); // Macro
  }
  p_sip_url.components() = uri_components;
  // UrlParameters
  OPTIONAL<LibSip__Common::SemicolonParam__List> params;
  fill_semi_colon_params(p_uri->url_params, params);
  p_sip_url.urlParameters() = params;
  // Headers
  OPTIONAL<LibSip__Common::AmpersandParam__List> headers;
  fill_ampersand_params(p_uri->url_headers, headers);
  p_sip_url.headers() = headers;
  
  loggers::get_instance().log_msg("<<< sip_codec_request::fill_uri: ", p_sip_url);
} // End of method fill_uri

void sip_codec_request::fill_authorization_header(const osip_message_t* p_sip_message, LibSip__SIPTypesAndValues::Authorization& p_authorization_header)
{
  loggers::get_instance().log(">>> sip_codec_request::fill_authorization_header: %p", p_sip_message->authorizations);
  
  // FieldName
  p_authorization_header.fieldName() = LibSip__SIPTypesAndValues::FieldName(LibSip__SIPTypesAndValues::FieldName::str_to_enum("AUTHORIZATION_E"));
  // CredentialsList
  LibSip__SIPTypesAndValues::CredentialsList l;
  int i = 0;
  osip_authorization_t *sip_authorization = nullptr;
  while(::osip_message_get_authorization(p_sip_message, i, &sip_authorization) == 0) {
    loggers::get_instance().log("sip_codec_request::fill_authorization_header: Processing item #%d", i);
    LibSip__SIPTypesAndValues::Credentials v;
    LibSip__Common::CommaParam__List cl;
    int j = 0;
    if (::osip_authorization_get_digest(sip_authorization) != nullptr) {
      LibSip__Common::GenValue v;
      v.tokenOrHost() = CHARSTRING(::osip_authorization_get_digest(sip_authorization));
      cl[j++] = LibSip__Common::GenericParam(
                                            CHARSTRING("Digest"),
                                            OPTIONAL<LibSip__Common::GenValue>(v)
                                            );
    }
    if (::osip_authorization_get_realm(sip_authorization) != nullptr) {
      LibSip__Common::GenValue v;
      v.tokenOrHost() = CHARSTRING(::osip_authorization_get_realm(sip_authorization));
      cl[j++] = LibSip__Common::GenericParam(
                                            CHARSTRING("Realm"),
                                            OPTIONAL<LibSip__Common::GenValue>(v)
                                            );
    }
    // FIXME To be contiuened
    v.digestResponse() = cl;
    l[i++] = v;
  } // End of 'while' statement
  p_authorization_header.body() = l;
  
  loggers::get_instance().log_msg("<<< sip_codec_request::fill_authorization_header: ", p_authorization_header);
} // End of method fill_authorization_header














void sip_codec_request::fill_from_header(const osip_from_t* p_sip_from, LibSip__SIPTypesAndValues::From& p_from_header)
{
  loggers::get_instance().log("sip_codec_request::fill_from_header");
  
  // FieldName
  p_from_header.fieldName() = LibSip__SIPTypesAndValues::FieldName(LibSip__SIPTypesAndValues::FieldName::str_to_enum("FROM_E"));
  // Addr_Union
  LibSip__SIPTypesAndValues::SipUrl uri;
  fill_uri(uri, ::osip_from_get_url((osip_from_t*)p_sip_from));
  OPTIONAL<LibSip__SIPTypesAndValues::DisplayName> display_name;
  if (::osip_from_get_displayname((osip_from_t*)p_sip_from) != nullptr) {
    LibSip__SIPTypesAndValues::DisplayName n;
    n.token() = CHARSTRING(::osip_from_get_displayname((osip_from_t*)p_sip_from));
    display_name = OPTIONAL<LibSip__SIPTypesAndValues::DisplayName>(n);
  } else {
    display_name.set_to_omit();
  }
  LibSip__SIPTypesAndValues::NameAddr name_addr(display_name, uri);
  LibSip__SIPTypesAndValues::Addr__Union addr;
  addr.nameAddr() = name_addr;
  p_from_header.addressField() = addr;
  // Params
  OPTIONAL<LibSip__Common::SemicolonParam__List> params;
  fill_semi_colon_params(p_sip_from->gen_params, params);
  p_from_header.fromParams() = params;
  
  loggers::get_instance().log_msg("<<< sip_codec_request::fill_from_header: ", p_from_header);
} // End of method fill_from_header

void sip_codec_request::fill_to_header(const osip_to_t* p_sip_to, LibSip__SIPTypesAndValues::To& p_to_header)
{
  loggers::get_instance().log("sip_codec_request::fill_to_header");
  
  // FieldName
  p_to_header.fieldName() = LibSip__SIPTypesAndValues::FieldName(LibSip__SIPTypesAndValues::FieldName::str_to_enum("TO_E"));
  // Addr_Union
  LibSip__SIPTypesAndValues::SipUrl uri;
  fill_uri(uri, ::osip_to_get_url((osip_to_t*)p_sip_to));
  OPTIONAL<LibSip__SIPTypesAndValues::DisplayName> display_name;
  if (::osip_to_get_displayname((osip_to_t*)p_sip_to) != nullptr) {
    LibSip__SIPTypesAndValues::DisplayName n;
    n.token() = CHARSTRING(::osip_to_get_displayname((osip_to_t*)p_sip_to));
    display_name = OPTIONAL<LibSip__SIPTypesAndValues::DisplayName>(n);
  } else {
    display_name.set_to_omit();
  }
  LibSip__SIPTypesAndValues::NameAddr name_addr(display_name, uri);
  LibSip__SIPTypesAndValues::Addr__Union addr;
  addr.nameAddr() = name_addr;
  p_to_header.addressField() = addr;
  // Params
  OPTIONAL<LibSip__Common::SemicolonParam__List> params;
  fill_semi_colon_params(p_sip_to->gen_params, params);
  p_to_header.toParams() = params;
  
  loggers::get_instance().log_msg("<<< sip_codec_request::fill_to_header: ", p_to_header);
} // End of method fill_to_header

void sip_codec_request::fill_accept_header(const osip_message_t* p_sip_message, LibSip__SIPTypesAndValues::Accept& p_accept_header)
{
  loggers::get_instance().log(">>> sip_codec_request::fill_accept_header: %p", p_sip_message->vias);

  // FieldName
  p_accept_header.fieldName() = LibSip__SIPTypesAndValues::FieldName(LibSip__SIPTypesAndValues::FieldName::str_to_enum("ACCEPT_E"));
  // AcceptArgs
  // TODO To be done
  p_accept_header.acceptArgs().set_to_omit();

  loggers::get_instance().log_msg("<<< sip_codec_request::fill_accept_header: ", p_accept_header);
} // End of method fill_accept_header

void sip_codec_request::fill_accept_contact_header(const osip_message_t* p_sip_message, LibSip__SIPTypesAndValues::AcceptContact& p_accept_contact_header)
{
  loggers::get_instance().log(">>> sip_codec_request::fill_accept_contact_header: %p", p_sip_message->vias);

  // FieldName
  p_accept_contact_header.fieldName() = LibSip__SIPTypesAndValues::FieldName(LibSip__SIPTypesAndValues::FieldName::str_to_enum("ACCEPT_CONTACT_E"));
  // AcRcValue
  LibSip__SIPTypesAndValues::AcRcValue__List l;
  // TODO To be done
  l[0] = LibSip__SIPTypesAndValues::AcRcValue(CHARSTRING("*"), OPTIONAL<LibSip__Common::SemicolonParam__List>(OMIT_VALUE));
  p_accept_contact_header.acValues() = l;

  loggers::get_instance().log_msg("<<< sip_codec_request::fill_accept_contact_header: ", p_accept_contact_header);
} // End of method fill_accept_contact_header

void sip_codec_request::fill_accept_encoding_header(const osip_message_t* p_sip_message, LibSip__SIPTypesAndValues::AcceptEncoding& p_accept_encoding_header)
{
  loggers::get_instance().log(">>> sip_codec_request::fill_accept_encoding_header: %p", p_sip_message->vias);

  // FieldName
  p_accept_encoding_header.fieldName() = LibSip__SIPTypesAndValues::FieldName(LibSip__SIPTypesAndValues::FieldName::str_to_enum("ACCEPT_ENCODING_E"));
  // ContentCoding
  p_accept_encoding_header.contentCoding().set_to_omit();

  loggers::get_instance().log_msg("<<< sip_codec_request::fill_accept_encoding_header: ", p_accept_encoding_header);
} // End of method fill_accept_encoding_header

void sip_codec_request::fill_accept_language_header(const osip_message_t* p_sip_message, LibSip__SIPTypesAndValues::AcceptLanguage& p_accept_language_header)
{
  loggers::get_instance().log(">>> sip_codec_request::fill_accept_language_header: %p", p_sip_message->vias);

  // FieldName
  p_accept_language_header.fieldName() = LibSip__SIPTypesAndValues::FieldName(LibSip__SIPTypesAndValues::FieldName::str_to_enum("ACCEPT_LANGUAGE_E"));
  // LanguageBody
  p_accept_language_header.languageBody().set_to_omit();

  loggers::get_instance().log_msg("<<< sip_codec_request::fill_accept_language_header: ", p_accept_language_header);
} // End of method fill_accept_language_header

void sip_codec_request::fill_alert_info_header(const osip_message_t* p_sip_message, LibSip__SIPTypesAndValues::AlertInfo& p_alert_info_header)
{
  loggers::get_instance().log(">>> sip_codec_request::fill_alert_info_header: %p", p_sip_message->vias);

  // FieldName
  p_alert_info_header.fieldName() = LibSip__SIPTypesAndValues::FieldName(LibSip__SIPTypesAndValues::FieldName::str_to_enum("ALERT_INFO_E"));
  // AlertInfoBody
  p_alert_info_header.alertInfoBody().set_to_omit();

  loggers::get_instance().log_msg("<<< sip_codec_request::fill_alert_info_header: ", p_alert_info_header);
} // End of method fill_alert_info_header

void sip_codec_request::fill_allow_header(const osip_message_t* p_sip_message, LibSip__SIPTypesAndValues::Allow& p_allow_header)
{
  loggers::get_instance().log(">>> sip_codec_request::fill_allow_header: %p", p_sip_message->vias);

  // FieldName
  p_allow_header.fieldName() = LibSip__SIPTypesAndValues::FieldName(LibSip__SIPTypesAndValues::FieldName::str_to_enum("ALLOW_E"));
  // Methods
  p_allow_header.methods().set_to_omit();

  loggers::get_instance().log_msg("<<< sip_codec_request::fill_allow_header: ", p_allow_header);
} // End of method fill_allow_header

void sip_codec_request::fill_allow_events_header(const osip_message_t* p_sip_message, LibSip__SIPTypesAndValues::AllowEvents& p_allow_events_header)
{
  loggers::get_instance().log(">>> sip_codec_request::fill_allow_events_header: %p", p_sip_message->vias);

  // FieldName
  p_allow_events_header.fieldName() = LibSip__SIPTypesAndValues::FieldName(LibSip__SIPTypesAndValues::FieldName::str_to_enum("ALLOW_EVENTS_E"));
  // EventTypes
  LibSip__SIPTypesAndValues::EventType__List l;
  // TODO To be done
  l[0] = CHARSTRING("*");
  p_allow_events_header.eventTypes() = l;

  loggers::get_instance().log_msg("<<< sip_codec_request::fill_allow_events_header: ", p_allow_events_header);
} // End of method fill_allow_events_header

void sip_codec_request::fill_via_header(const osip_message_t* p_sip_message, LibSip__SIPTypesAndValues::Via& p_via_header)
{
  loggers::get_instance().log(">>> sip_codec_request::fill_via_header: %p", p_sip_message->vias);
  
  // FieldName
  p_via_header.fieldName() = LibSip__SIPTypesAndValues::FieldName(LibSip__SIPTypesAndValues::FieldName::str_to_enum("VIA_E"));
  // ViaBody
  LibSip__SIPTypesAndValues::ViaBody__List l;
  int i = 0;
  osip_via_t *sip_via = nullptr;
  while(::osip_message_get_via(p_sip_message, i, &sip_via) == 0) {
    loggers::get_instance().log("sip_codec_request::fill_via_header: Processing item #%d", i);
    LibSip__SIPTypesAndValues::ViaBody v;
    v.sentProtocol() = LibSip__SIPTypesAndValues::SentProtocol(
                                                               CHARSTRING(::via_get_protocol(sip_via)),
                                                               CHARSTRING(::via_get_version(sip_via)),
                                                               CHARSTRING(::via_get_comment(sip_via))
                                                               );
    OPTIONAL<LibSip__SIPTypesAndValues::HostPort> host_port;
    fill_host_port(::via_get_host(sip_via), ::via_get_port(sip_via), host_port);
    loggers::get_instance().log_msg("sip_codec_request::fill_via_uri: host_port: ", host_port);
    v.sentBy() = host_port;
    OPTIONAL<LibSip__Common::SemicolonParam__List> params;
    fill_semi_colon_params(sip_via->via_params, params);
    v.viaParams() = params;
    l[i++] = v;
  } // End of 'while' statement
  p_via_header.viaBody() = l;
  
  loggers::get_instance().log_msg("<<< sip_codec_request::fill_via_header: ", p_via_header);
} // End of method fill_via_header

/***

    inline OPTIONAL<AuthenticationInfo>& authenticationInfo()
    {return field_authenticationInfo;}
    inline const OPTIONAL<AuthenticationInfo>& authenticationInfo() const
    {return field_authenticationInfo;}
    inline OPTIONAL<Authorization>& authorization()
    {return field_authorization;}
    inline const OPTIONAL<Authorization>& authorization() const
    {return field_authorization;}
    inline OPTIONAL<CallId>& callId()
    {return field_callId;}
    inline const OPTIONAL<CallId>& callId() const
    {return field_callId;}
    inline OPTIONAL<CallInfo>& callInfo()
    {return field_callInfo;}
    inline const OPTIONAL<CallInfo>& callInfo() const
    {return field_callInfo;}
    inline OPTIONAL<Contact>& contact()
    {return field_contact;}
    inline const OPTIONAL<Contact>& contact() const
    {return field_contact;}
    inline OPTIONAL<ContentDisposition>& contentDisposition()
    {return field_contentDisposition;}
    inline const OPTIONAL<ContentDisposition>& contentDisposition() const
    {return field_contentDisposition;}
    inline OPTIONAL<ContentEncoding>& contentEncoding()
    {return field_contentEncoding;}
    inline const OPTIONAL<ContentEncoding>& contentEncoding() const
    {return field_contentEncoding;}
    inline OPTIONAL<ContentLanguage>& contentLanguage()
    {return field_contentLanguage;}
    inline const OPTIONAL<ContentLanguage>& contentLanguage() const
    {return field_contentLanguage;}
    inline OPTIONAL<ContentLength>& contentLength()
    {return field_contentLength;}
    inline const OPTIONAL<ContentLength>& contentLength() const
    {return field_contentLength;}
    inline OPTIONAL<ContentType>& contentType()
    {return field_contentType;}
    inline const OPTIONAL<ContentType>& contentType() const
    {return field_contentType;}
    inline OPTIONAL<CSeq>& cSeq()
    {return field_cSeq;}
    inline const OPTIONAL<CSeq>& cSeq() const
    {return field_cSeq;}
    inline OPTIONAL<Date>& date()
    {return field_date;}
    inline const OPTIONAL<Date>& date() const
    {return field_date;}
    inline OPTIONAL<Diversion>& diversion()
    {return field_diversion;}
    inline const OPTIONAL<Diversion>& diversion() const
    {return field_diversion;}
    inline OPTIONAL<ErrorInfo>& errorInfo()
    {return field_errorInfo;}
    inline const OPTIONAL<ErrorInfo>& errorInfo() const
    {return field_errorInfo;}
    inline OPTIONAL<Event>& event()
    {return field_event;}
    inline const OPTIONAL<Event>& event() const
    {return field_event;}
    inline OPTIONAL<Expires>& expires()
    {return field_expires;}
    inline const OPTIONAL<Expires>& expires() const
    {return field_expires;}
    inline OPTIONAL<FeatureCaps>& featureCaps()
    {return field_featureCaps;}
    inline const OPTIONAL<FeatureCaps>& featureCaps() const
    {return field_featureCaps;}
    inline OPTIONAL<Geolocation>& geolocation()
    {return field_geolocation;}
    inline const OPTIONAL<Geolocation>& geolocation() const
    {return field_geolocation;}
    inline OPTIONAL<GeolocationRouting>& geolocationRouting()
    {return field_geolocationRouting;}
    inline const OPTIONAL<GeolocationRouting>& geolocationRouting() const
    {return field_geolocationRouting;}
    inline OPTIONAL<HistoryInfo>& historyInfo()
    {return field_historyInfo;}
    inline const OPTIONAL<HistoryInfo>& historyInfo() const
    {return field_historyInfo;}
    inline OPTIONAL<InfoPackage>& infoPackage()
    {return field_infoPackage;}
    inline const OPTIONAL<InfoPackage>& infoPackage() const
    {return field_infoPackage;}
    inline OPTIONAL<InReplyTo>& inReplyTo()
    {return field_inReplyTo;}
    inline const OPTIONAL<InReplyTo>& inReplyTo() const
    {return field_inReplyTo;}
    inline OPTIONAL<MaxForwards>& maxForwards()
    {return field_maxForwards;}
    inline const OPTIONAL<MaxForwards>& maxForwards() const
    {return field_maxForwards;}
    inline OPTIONAL<MimeVersion>& mimeVersion()
    {return field_mimeVersion;}
    inline const OPTIONAL<MimeVersion>& mimeVersion() const
    {return field_mimeVersion;}
    inline OPTIONAL<MinExpires>& minExpires()
    {return field_minExpires;}
    inline const OPTIONAL<MinExpires>& minExpires() const
    {return field_minExpires;}
    inline OPTIONAL<MinSE>& minSE()
    {return field_minSE;}
    inline const OPTIONAL<MinSE>& minSE() const
    {return field_minSE;}
    inline OPTIONAL<Organization>& organization()
    {return field_organization;}
    inline const OPTIONAL<Organization>& organization() const
    {return field_organization;}
    inline OPTIONAL<PAccessNetworkInfo>& pAccessNetworkInfo()
    {return field_pAccessNetworkInfo;}
    inline const OPTIONAL<PAccessNetworkInfo>& pAccessNetworkInfo() const
    {return field_pAccessNetworkInfo;}
    inline OPTIONAL<PAssertedID>& pAssertedID()
    {return field_pAssertedID;}
    inline const OPTIONAL<PAssertedID>& pAssertedID() const
    {return field_pAssertedID;}
    inline OPTIONAL<PAssertedService>& pAssertedService()
    {return field_pAssertedService;}
    inline const OPTIONAL<PAssertedService>& pAssertedService() const
    {return field_pAssertedService;}
    inline OPTIONAL<PAssociatedURI>& pAssociatedURI()
    {return field_pAssociatedURI;}
    inline const OPTIONAL<PAssociatedURI>& pAssociatedURI() const
    {return field_pAssociatedURI;}
    inline OPTIONAL<Path>& path()
    {return field_path;}
    inline const OPTIONAL<Path>& path() const
    {return field_path;}
    inline OPTIONAL<PCalledPartyID>& pCalledPartyID()
    {return field_pCalledPartyID;}
    inline const OPTIONAL<PCalledPartyID>& pCalledPartyID() const
    {return field_pCalledPartyID;}
    inline OPTIONAL<PChargingFunctionAddresses>& pChargingFunctionAddresses()
    {return field_pChargingFunctionAddresses;}
    inline const OPTIONAL<PChargingFunctionAddresses>& pChargingFunctionAddresses() const
    {return field_pChargingFunctionAddresses;}
    inline OPTIONAL<PChargingVector>& pChargingVector()
    {return field_pChargingVector;}
    inline const OPTIONAL<PChargingVector>& pChargingVector() const
    {return field_pChargingVector;}
    inline OPTIONAL<PEarlyMedia>& pEarlyMedia()
    {return field_pEarlyMedia;}
    inline const OPTIONAL<PEarlyMedia>& pEarlyMedia() const
    {return field_pEarlyMedia;}
    inline OPTIONAL<PMediaAuthorization>& pMediaAuthorization()
    {return field_pMediaAuthorization;}
    inline const OPTIONAL<PMediaAuthorization>& pMediaAuthorization() const
    {return field_pMediaAuthorization;}
    inline OPTIONAL<PPreferredID>& pPreferredID()
    {return field_pPreferredID;}
    inline const OPTIONAL<PPreferredID>& pPreferredID() const
    {return field_pPreferredID;}
    inline OPTIONAL<PPreferredService>& pPreferredService()
    {return field_pPreferredService;}
    inline const OPTIONAL<PPreferredService>& pPreferredService() const
    {return field_pPreferredService;}
    inline OPTIONAL<Priority>& priority()
    {return field_priority;}
    inline const OPTIONAL<Priority>& priority() const
    {return field_priority;}
    inline OPTIONAL<Privacy>& privacy()
    {return field_privacy;}
    inline const OPTIONAL<Privacy>& privacy() const
    {return field_privacy;}
    inline OPTIONAL<ProxyAuthenticate>& proxyAuthenticate()
    {return field_proxyAuthenticate;}
    inline const OPTIONAL<ProxyAuthenticate>& proxyAuthenticate() const
    {return field_proxyAuthenticate;}
    inline OPTIONAL<ProxyAuthorization>& proxyAuthorization()
    {return field_proxyAuthorization;}
    inline const OPTIONAL<ProxyAuthorization>& proxyAuthorization() const
    {return field_proxyAuthorization;}
    inline OPTIONAL<ProxyRequire>& proxyRequire()
    {return field_proxyRequire;}
    inline const OPTIONAL<ProxyRequire>& proxyRequire() const
    {return field_proxyRequire;}
    inline OPTIONAL<PVisitedNetworkID>& pVisitedNetworkID()
    {return field_pVisitedNetworkID;}
    inline const OPTIONAL<PVisitedNetworkID>& pVisitedNetworkID() const
    {return field_pVisitedNetworkID;}
    inline OPTIONAL<RAck>& rAck()
    {return field_rAck;}
    inline const OPTIONAL<RAck>& rAck() const
    {return field_rAck;}
    inline OPTIONAL<RSeq>& rSeq()
    {return field_rSeq;}
    inline const OPTIONAL<RSeq>& rSeq() const
    {return field_rSeq;}
    inline OPTIONAL<Reason>& reason()
    {return field_reason;}
    inline const OPTIONAL<Reason>& reason() const
    {return field_reason;}
    inline OPTIONAL<RecordRoute>& recordRoute()
    {return field_recordRoute;}
    inline const OPTIONAL<RecordRoute>& recordRoute() const
    {return field_recordRoute;}
    inline OPTIONAL<RecvInfo>& recvInfo()
    {return field_recvInfo;}
    inline const OPTIONAL<RecvInfo>& recvInfo() const
    {return field_recvInfo;}
    inline OPTIONAL<RequestDisposition>& requestDisposition()
    {return field_requestDisposition;}
    inline const OPTIONAL<RequestDisposition>& requestDisposition() const
    {return field_requestDisposition;}
    inline OPTIONAL<ReferredBy>& referredBy()
    {return field_referredBy;}
    inline const OPTIONAL<ReferredBy>& referredBy() const
    {return field_referredBy;}
    inline OPTIONAL<ReferTo>& referTo()
    {return field_referTo;}
    inline const OPTIONAL<ReferTo>& referTo() const
    {return field_referTo;}
    inline OPTIONAL<ReferSub>& referSub()
    {return field_referSub;}
    inline const OPTIONAL<ReferSub>& referSub() const
    {return field_referSub;}
    inline OPTIONAL<Replaces>& replaces()
    {return field_replaces;}
    inline const OPTIONAL<Replaces>& replaces() const
    {return field_replaces;}
    inline OPTIONAL<ReplyTo>& replyTo()
    {return field_replyTo;}
    inline const OPTIONAL<ReplyTo>& replyTo() const
    {return field_replyTo;}
    inline OPTIONAL<Require>& require()
    {return field_require;}
    inline const OPTIONAL<Require>& require() const
    {return field_require;}
    inline OPTIONAL<RetryAfter>& retryAfter()
    {return field_retryAfter;}
    inline const OPTIONAL<RetryAfter>& retryAfter() const
    {return field_retryAfter;}
    inline OPTIONAL<Route>& route()
    {return field_route;}
    inline const OPTIONAL<Route>& route() const
    {return field_route;}
    inline OPTIONAL<SecurityClient>& securityClient()
    {return field_securityClient;}
    inline const OPTIONAL<SecurityClient>& securityClient() const
    {return field_securityClient;}
    inline OPTIONAL<SecurityServer>& securityServer()
    {return field_securityServer;}
    inline const OPTIONAL<SecurityServer>& securityServer() const
    {return field_securityServer;}
    inline OPTIONAL<SecurityVerify>& securityVerify()
    {return field_securityVerify;}
    inline const OPTIONAL<SecurityVerify>& securityVerify() const
    {return field_securityVerify;}
    inline OPTIONAL<Server>& server()
    {return field_server;}
    inline const OPTIONAL<Server>& server() const
    {return field_server;}
    inline OPTIONAL<ServiceRoute>& serviceRoute()
    {return field_serviceRoute;}
    inline const OPTIONAL<ServiceRoute>& serviceRoute() const
    {return field_serviceRoute;}
    inline OPTIONAL<SessionExpires>& sessionExpires()
    {return field_sessionExpires;}
    inline const OPTIONAL<SessionExpires>& sessionExpires() const
    {return field_sessionExpires;}
    inline OPTIONAL<SessionId>& sessionId()
    {return field_sessionId;}
    inline const OPTIONAL<SessionId>& sessionId() const
    {return field_sessionId;}
    inline OPTIONAL<SIP__ETag>& sipETag()
    {return field_sipETag;}
    inline const OPTIONAL<SIP__ETag>& sipETag() const
    {return field_sipETag;}
    inline OPTIONAL<SIP__If__Match>& sipIfMatch()
    {return field_sipIfMatch;}
    inline const OPTIONAL<SIP__If__Match>& sipIfMatch() const
    {return field_sipIfMatch;}
    inline OPTIONAL<Subject>& subject()
    {return field_subject;}
    inline const OPTIONAL<Subject>& subject() const
    {return field_subject;}
    inline OPTIONAL<SubscriptionState>& subscriptionState()
    {return field_subscriptionState;}
    inline const OPTIONAL<SubscriptionState>& subscriptionState() const
    {return field_subscriptionState;}
    inline OPTIONAL<Supported>& supported()
    {return field_supported;}
    inline const OPTIONAL<Supported>& supported() const
    {return field_supported;}
    inline OPTIONAL<Timestamp>& timestamp()
    {return field_timestamp;}
    inline const OPTIONAL<Timestamp>& timestamp() const
    {return field_timestamp;}
    inline OPTIONAL<Unsupported>& unsupported()
    {return field_unsupported;}
    inline const OPTIONAL<Unsupported>& unsupported() const
    {return field_unsupported;}
    inline OPTIONAL<UserToUser>& userToUser()
    {return field_userToUser;}
    inline const OPTIONAL<UserToUser>& userToUser() const
    {return field_userToUser;}
    inline OPTIONAL<UserAgent>& userAgent()
    {return field_userAgent;}
    inline const OPTIONAL<UserAgent>& userAgent() const
    {return field_userAgent;}
    inline OPTIONAL<Warning>& warning()
    {return field_warning;}
    inline const OPTIONAL<Warning>& warning() const
    {return field_warning;}
    inline OPTIONAL<WwwAuthenticate>& wwwAuthenticate()
    {return field_wwwAuthenticate;}
    inline const OPTIONAL<WwwAuthenticate>& wwwAuthenticate() const
    {return field_wwwAuthenticate;}
    inline OPTIONAL<ResourcePriority>& resourcePriority()
    {return field_resourcePriority;}
    inline const OPTIONAL<ResourcePriority>& resourcePriority() const
    {return field_resourcePriority;}
    inline OPTIONAL<AnswerMode>& answerMode()
    {return field_answerMode;}
    inline const OPTIONAL<AnswerMode>& answerMode() const
    {return field_answerMode;}
    inline OPTIONAL<PrivAnswerMode>& privAnswerMode()
    {return field_privAnswerMode;}
    inline const OPTIONAL<PrivAnswerMode>& privAnswerMode() const
    {return field_privAnswerMode;}
    inline OPTIONAL<TargetDialog>& targetDialog()
    {return field_targetDialog;}
    inline const OPTIONAL<TargetDialog>& targetDialog() const
    {return field_targetDialog;}
    inline OPTIONAL<PAnswerState>& pAnswerState()
    {return field_pAnswerState;}
    inline const OPTIONAL<PAnswerState>& pAnswerState() const
    {return field_pAnswerState;}
    inline OPTIONAL<UndefinedHeader__List>& undefinedHeader__List()
    {return field_undefinedHeader__List;}
    inline const OPTIONAL<UndefinedHeader__List>& undefinedHeader__List() const
    {return field_undefinedHeader__List;}

***/
